/*
 * File: rho.h
 * Date: 2011-11-01
 *
 * Evsiz (evsiz@this-voice.org)
 * Rho released under the Creative Commons NonCommercial ShareAlike License 3.0
 * See http://creativecommons.org for details
 *
 * Rho is a fork of the Linux port of MicroEMACS developed by Humberto
 * Ortiz-Zuazaga and Jean-Pierre Demailly. Their work forked the MicroEMACS
 * of Daniel M. Lawrence which was in turn a fork of uemacs by Dave G.
 * Conroy. MicroEmacs was copyrighted 1993 by Mr. Lawrence with a
 * stipulation that it could not be used commercially. Uemacs was released
 * by Mr. Conroy into the public domain in 1985. This file based on code by
 * Dave G. Conroy, Steve Wilhite and George Jones.
 *
 * MACOS port by Aria Stewart <aredridel@nbtsc.org>
 */

#define VERSION   "1.1.14"


/*** ------------------- BEGIN USER DEFINED OPTIONS -----------------***/

/* PORTS - POSIX is default
 *   See README for notes on broken ports
 */

#define MACOS   0     /* for OSX and possibly other BSDs */


/*  Make your TERM behave properly  */

#define SHELLINIT "stty -ixon" /* Add IXON, IXANY, IXOFF ? */


/*  Language text options */

#define ENGLISH  1
#define FRENCH   0
#define SPANISH  0
#define GERMAN   0
#define DUTCH    0
#define JAPANESE 0

/*  Height of Pop-up Windows */

#define POPUPS   8  /* most popups    */
#define ERRWIN   6   /* compile errors */

/*  Directories (See rho_init.h and Makefile)   */

#define BINDIR    "/usr/local/bin/"
#define SHRDIR    "/usr/local/share/rho/"


/*** ------------------- END USER DEFINED OPTIONS -----------------***/

#if ENGLISH
#include  "lang_en.h"
#elif FRENCH
#include  "lang_fr.h"
#elif SPANISH
#include  "lang_es.h"
#elif GERMAN
#include  "lang_de.h"
#elif DUTCH
#include  "lang_nl.h"
#elif JAPANESE
#include  "lang_ja.h"
#endif


/*  miscellaneous settings  */

#define DIRSEPCHAR  '/'
#define DIRSEPSTR "/"
#define PATHCHR   ':'
#define PROGNAME  "Rho"
#define RETCHAR   '\r'
#define RHO_RC    ".rhorc"

#define RHOOUT  "_rho_stdout"    /* stdout file for callout functions */
#define RHOERR  "_rho_stderr"    /* stderr file for callout functions */


/*  internal macros   */

#define nextab(a) (a - (a % hardtab)) + hardtab
#define NUFUNCS sizeof(ufuncs) / sizeof(UFUNC)


/*  internal constants  */


#define GFREAD     1   /* read first file in at startup (for gflags)   */
#define GFSDRAW    2   /* suppress a screen redraw (for gflags)        */
#define HUGE    1000   /* Huge number (kinda)          */
#define KBLOCK   250   /* sizeof kill buffer chunks    */
#define LENPATH  128   /* matches bytes in path name   */
#define NBINDS   300   /* max # of bound keys          */
#define NBLOCK    16   /* line block chunk size        */
#define NKBDM    256   /* # of strokes, keyboard macro */
#define NLINE    256   /* # of bytes, input line       */
#define NMARKS    16   /* number of marks              */
#define NPAT     128   /* # of bytes, pattern          */
#define NRING     16   /* # of buffers in kill ring    */
#define NSTRING  128   /* # of bytes, string buffers   */
#define NVSIZE    10   /* max #chars in a var name     */

#ifdef  CTRL
#undef  CTRL
#endif
#define CTRL  0x0100    /* Control flag, or'ed in      */
#define META  0x0200    /* Meta flag, or'ed in         */
#define CTLX  0x0400    /* ^X flag, or'ed in           */
#define SPEC  0x0800    /* special key (function keys) */
#define SHFT  0x2000    /* shifted (for function keys) */

#define BINDNUL 0   /* not bound to anything           */
#define BINDFNC 1   /* key bound to a function         */
#define BINDBUF 2   /* key bound to a buffer           */

#ifdef  FALSE
#undef  FALSE
#endif
#ifdef  TRUE
#undef  TRUE
#endif

#define FALSE 0
#define TRUE  1
#define ABORT 2

#define STOP    0     /* keyboard macro not in use    */
#define PLAY    1     /*    "       "   playing       */
#define RECORD  2     /*    "       "   recording     */

/*  Completion types  */

#define CMP_BUFFER    0
#define CMP_COMMAND   1
#define CMP_FILENAME  2

/*  Directive definitions */

#define DIF       0
#define DELSE     1
#define DENDIF    2
#define DGOTO     3
#define DRETURN   4
#define DENDM     5
#define DWHILE    6
#define DENDWHILE 7
#define DBREAK    8
#define DFORCE    9
#define NUMDIRS   10  /* total directives */

/*
 * PTBEG, PTEND, FORWARD, and REVERSE are all toggle-able values for
 * routines that need directions.
 */
#define PTBEG   0   /* Leave the point at the beginning on search   */
#define PTEND   1   /* Leave the point at the end on search         */
#define FORWARD 0   /* do things in a forward direction             */
#define REVERSE 1   /* do things in a backwards direction           */

#define FIOSUC  0   /* File I/O, success.           */
#define FIOFNF  1   /* File I/O, file not found.    */
#define FIOEOF  2   /* File I/O, end of file.       */
#define FIOERR  3   /* File I/O, error.             */
#define FIOMEM  4   /* File I/O, out of memory      */

#define CFCPCN  0x0001    /* Last command was C-P, C-N    */
#define CFKILL  0x0002    /* Last command was a kill      */
#define CFSRCH  0x0004    /* last command was a search    */
#define CFYANK  0x0008    /* last command was a yank      */

#define SRNORM  0   /* end past, begin front        */
#define SRBEGIN 1   /* always at front              */

#define BELL  0x07  /* a bell character             */

/*  Macro argument token types          */

#define TKNUL 0   /* end-of-string                */
#define TKARG 1   /* interactive argument         */
#define TKBUF 2   /* buffer argument              */
#define TKVAR 3   /* user variables               */
#define TKENV 4   /* environment variables        */
#define TKFUN 5   /* function....                 */
#define TKDIR 6   /* directive                    */
#define TKLBL 7   /* line label                   */
#define TKLIT 8   /* numeric literal              */
#define TKSTR 9   /* quoted string literal        */
#define TKCMD 10  /* command name                 */

/*
 * There is a window structure allocated for every active display window. The
 * windows are kept in a big list, in top to bottom screen order, with the
 * listhead at "wheadp". Each window contains its own values of dot and mark.
 * The flag field contains some bits that are set by commands to guide
 * redisplay. Although this is a bit of a compromise in terms of decoupling,
 * the full blown redisplay is just too expensive to run for every input
 * character.
 */
typedef struct WNDW
{
  struct WNDW *w_wndp;          /* Next window                  */
  struct BUFFER *w_bufp;          /* Buffer displayed in window   */
  struct LINE *w_linep;           /* Top line in the window       */
  struct LINE *w_dotp;            /* Line containing "."          */
  int w_doto;                     /* Byte offset for "."          */
  struct LINE *w_markp[NMARKS];   /* Line containing "mark"       */
  int w_marko[NMARKS];            /* Byte offset for "mark"       */
  char w_toprow;                  /* Origin 0 top row of window   */
  char w_ntrows;                  /* # of rows of text in window  */
  char w_force;                   /* If NZ, forcing row.          */
  char w_flag;                    /* Flags.                       */
  int w_fcol;                     /* first column displayed       */
} WNDW;

#define WFFORCE 0x01    /* Window needs forced reframe  */
#define WFMOVE  0x02    /* Movement from line to line   */
#define WFEDIT  0x04    /* Editing within a line        */
#define WFHARD  0x08    /* Better update full display   */
#define WFMODE  0x10    /* Update mode line.            */

/* This structure holds the information about each line appearing on the
 * video display. The redisplay module uses an array of virtual display
 * lines. On systems that do not have direct access to display memory,
 * there is also an array of physical display lines used to minimize
 * video updating. In most cases, these two arrays are unique. 
 */

#define FARRIGHT  999 /* column beyond the right edge! */

typedef struct VIDEO
{
  int v_flag;       /* Flags */
  int v_left;       /* left edge of reverse video */
  int v_right;      /* right right of reverse video */
  char v_text[1];   /* Screen data. */
} VIDEO;

#define VFNEW 0x0001    /* contents not meaningful yet  */
#define VFCHG 0x0002    /* Changed flag                 */
#define VFEXT 0x0004    /* extended (beyond column 80)  */
#define VFCOL 0x0008    /* color change requested       */


/*
 *  This structure holds the information about each separate "screen"
 * within the current editing session.  On a character based system, these
 * screens overlay each other, and can individually be brought to front.
 *
 */

typedef struct SCRN
{
  struct SCRN *s_next_screen; /* link to next screen in list */
  WNDW *s_first_window;       /* head of linked list of windows */
  WNDW *s_cur_window;         /* current window in this screen */
  char *s_screen_name;          /* name of the current window */
  int s_roworg;                 /* row origin of stored screen */
  int s_colorg;                 /* column origin of stored screen */
  int s_nrow;                   /* row width of stored screen */
  int s_ncol;                   /* column origin of stored screen */
} SCRN;

/*
 * Text is kept in buffers. A buffer header, described below, exists for every
 * buffer in the system. The buffers are kept in a big list, so that commands
 * that search for a buffer by name can find the buffer header. There is a
 * safe store for the dot and mark in the header, but this is only valid if
 * the buffer is not being displayed (that is, if "b_nwnd" is 0). The text for
 * the buffer is kept in a circularly linked list of lines, with a pointer to
 * the header line in "b_linep".  Buffers may be "Inactive" which means the
 * files associated with them have not been read in yet.  These get read in
 * at "use buffer" time.
 */
typedef struct BUFFER
{
  struct BUFFER *b_bufp;        /* Link to next BUFFER          */
  struct LINE *b_dotp;          /* Link to "." LINE structure   */
  int b_doto;                   /* Offset of "." in above LINE  */
  struct LINE *b_markp[NMARKS]; /* The same as the above two,   */
  int b_marko[NMARKS];          /* but for the "mark"           */
  int b_fcol;                   /* first col to display         */
  struct LINE *b_linep;         /* Link to the header LINE      */
  struct LINE *b_topline;       /* Link to narrowed top text    */
  struct LINE *b_botline;       /* Link to narrowed bottom text */
  char b_active;                /* window activated flag        */
  char b_nwnd;                  /* Count of windows on buffer   */
  char b_exec;                  /* how many active executions   */
  char b_flag;                  /* Flags                        */
  int b_mode;                   /* editor mode of this buffer   */
  char b_name[LENPATH];         /* File/Buffer name             */
} BUFFER;

#define BFINVS  0x01            /* Internal invisable buffer    */
#define BFCHG   0x02            /* Changed since last write     */
#define BFTRUNC 0x04            /* buffer was truncated when read */
#define BFNAROW 0x08            /* buffer has been narrowed     */

/*  mode flags  */
#define NUMMODES  10      /* # of defined modes           */

#define MDWRAP  0x0001    /* word wrap                    */
#define MDCMOD  0x0002    /* C indentation and fence match */
#define MDSPELL 0x0004    /* UNUSED */
#define MDEXACT 0x0008    /* Exact matching for searches  */
#define MDVIEW  0x0010    /* read-only buffer             */
#define MDOVER  0x0020    /* overwrite mode               */
#define MDMAGIC 0x0040    /* regular expresions in search */
#define MDCRYPT 0x0080    /* UNUSED */
#define MDASAVE 0x0100    /* auto-save mode               */
#define MDREPL  0x0200    /* replace mode                 */


/*
 * The starting position of a region, and the size of the region in
 * characters, is kept in a region structure.  Used by the region commands.
 */
 
typedef struct
{
  struct LINE *r_linep;   /* Origin LINE address.         */
  int r_offset;           /* Origin LINE offset.          */
  long r_size;            /* Length in characters.        */
} REGION;

/*
 * All text is kept in circularly linked lists of "LINE" structures. These
 * begin at the header line (which is the blank line beyond the end of the
 * buffer). This line is pointed to by the "BUFFER". Each line contains a the
 * number of bytes in the line (the "used" size), the size of the text array,
 * and the text. The end of line is not stored as a byte; it's implied.
 */
typedef struct LINE
{
  struct LINE *l_fp;    /* Link to the next line        */
  struct LINE *l_bp;    /* Link to the previous line    */
  size_t l_size;        /* Allocated size               */
  size_t l_used;        /* Used size                    */
  char l_text[1];       /* A bunch of characters.       */
} LINE;

#define lforw(lp) ((lp)->l_fp)
#define lback(lp) ((lp)->l_bp)
#define lgetc(lp, n)  ((lp)->l_text[(n)])
#define lputc(lp, n, c) ((lp)->l_text[(n)]=(c))
#define lused(lp) ((lp)->l_used)
#define ltext(lp) ((lp)->l_text)

/*
 * The editor communicates with the display using a high level interface. A
 * "TERM" structure holds useful variables, and indirect pointers to routines
 * that do useful operations. The low level get and put routines are here too.
 * This lets a terminal, in addition to having non standard commands, have
 * funny get and put character code too. The calls might get changed to
 * "termp->t_field" style in the future, to make it possible to run more than
 * one terminal type.
 */
typedef struct
{
  int t_mrow;               /* max number of rows allowable */
  int t_nrow;               /* current number of rows used  */
  int t_mcol;               /* max Number of columns.       */
  int t_ncol;               /* current Number of columns.   */
  int t_roworg;             /* origin row (normally zero)   */
  int t_colorg;             /* origin column (normally zero) */
  int t_margin;             /* min margin for extended lines */
  int t_scrsiz;             /* size of scroll region "      */
  int t_pause;              /* # times thru update to pause */
  int (*t_open) (void);     /* Open terminal at the start. */
  int (*t_close) (void);    /* Close terminal at end.       */
  int (*t_kopen) (void);    /* Open keyboard                */
  int (*t_kclose) (void);   /* Close keyboard               */
  int (*t_getchar) (void);  /* Get character from keyboard. */
  int (*t_putchar) (int);   /* Put character to display. */
  int (*t_flush) (void);    /* Flush output buffers.        */
  int (*t_move) (int, int); /* Move the cursor, origin 0. */
  int (*t_eeol) (void);     /* Erase to end of line.        */
  int (*t_eeop) (void);     /* Erase to end of page.        */
  int (*t_clrdesk) (void);  /* Clear the page totally       */
  int (*t_beep) (void);     /* Beep.                        */
  int (*t_rev) (int);       /* set reverse video state      */
  int (*t_rez) (char *);    /* change screen resolution     */
} TERM;


/*  macros for terminal I/O   */

#define TTbeep    (*term.t_beep)
#define TTclose   (*term.t_close)
#define TTeeol    (*term.t_eeol)
#define TTeeop    (*term.t_eeop)
#define TTflush   (*term.t_flush)
#define TTgetc    (*term.t_getchar)
#define TTkclose  (*term.t_kclose)
#define TTkopen   (*term.t_kopen)
#define TTmove    (*term.t_move)
#define TTopen    (*term.t_open)
#define TTputc    (*term.t_putchar)
#define TTrev     (*term.t_rev)

/*  Structure for the table of current key bindings   */


typedef struct
{
  int k_code;       /* Key code                     */
  int k_type;       /* binding type (C function or Rho buffer) */
  int (*k_fp) ();   /* C routine to invoke */
  BUFFER *k_buf;    /* buffer to execute */
} KEYTAB;

/*  structure for the name binding table    */

typedef struct
{
  char *n_name;       /* fn_string which is */
  int (*n_func) ();   /* bound to fn_name */
} NBIND;

/*  The editor holds deleted text chunks in the KILL buffer. The
  kill buffer is logically a stream of ascii characters, however
  due to its unpredicatable size, it gets implemented as a linked
  list of chunks. (The d_ prefix is for "deleted" text, as k_
  was taken up by the keycode structure)
*/

typedef struct KILL
{
  struct KILL *d_next;    /* link to next chunk, NULL if last */
  char d_chunk[KBLOCK];   /* deleted text */
} KILL;

/*  When rho's command interpeter needs to get a variable's name,
  rather than it's value, it is passed back as a VDESC variable
  description structure. The v_num field is a index into the
  appropriate variable table.
*/

typedef struct VDESC
{
  int v_type;     /* type of variable */
  int v_num;      /* ordinal pointer to variable in list */
} VDESC;

/*  The !WHILE directive in the execution language needs to
  stack references to pending whiles. These are stored linked
  to each currently open procedure via a linked list of
  the following structure
*/

typedef struct WHBLOCK
{
  LINE *w_begin;          /* ptr to !while statement */
  LINE *w_end;            /* ptr to the !endwhile statement */
  int w_type;             /* block type */
  struct WHBLOCK *w_next; /* next while */
} WHBLOCK;

#define BTWHILE   1
#define BTBREAK   2


/* HICHAR - 1 is the largest character we will deal with.
 * BMAPSIZE represents the number of bytes in the bitmap.
 */

#define HICHAR    256
#define BMAPSIZE  HICHAR >> 3

/*
 * Defines for the metacharacters in the regular expression
 * search routines.  MCNIL and GROUP are used in both search
 * and replace metachar-arrays.
 */
#define MCNIL     0    /* Like the '\0' for strings. */
#define LITSTRING 1    /* Literal string. */
#define LITCHAR   2    /* Literal character. */
#define ANY       3    /* Any character but the <NL>. */
#define CCL       4
#define NCCL      5
#define BOL       6
#define EOL       7
#define GRPBEG    8    /* Signal start of group. */
#define GRPEND    9    /* Signal end of group. */
#define GROUP    10    /* String of group match. */
#define DITTO    11    /* Replacement with match string. */

#define CLOSURE   0x0100  /* An or-able value for a closure modifier. */
#define CLOSURE_1 0x0200  /* An or-able value for a closure modifier. */
#define ZEROONE   0x0400  /* An or-able value for a closure modifier. */

#define ALLCLOS   (CLOSURE | CLOSURE_1 | ZEROONE)
#define MASKCLO   (~ALLCLOS)

#define MC_ANY    '.'      /* 'Any' character (except newline). */
#define MC_CCL    '['      /* Character class. */
#define MC_NCCL   '^'      /* Negate character class. */
#define MC_RCCL   '-'      /* Range in character class. */
#define MC_ECCL   ']'      /* End of character class. */
#define MC_BOL    '^'      /* Beginning of line. */
#define MC_EOL    '$'      /* End of line. */
#define MC_CLOSURE  '*'    /* Closure - zero to many characters match. */
#define MC_CLOSURE_1  '+'  /* Closure - one to many characters match. */
#define MC_ZEROONE  '?'    /* Closure - zero to one characters match. */
#define MC_DITTO  '&'      /* Use matched string in replacement. */
#define MC_GRPBEG '('      /* Start of group (begun with a backslash). */
#define MC_GRPEND ')'      /* End of group (begun with a backslash). */
#define MC_ESC    '\\'     /* Escape - suppress meta-meaning. */

#define MAXGROUPS 10          /* 1 + maximum # of r. e. groups. */
#define BIT(n)    (1 << (n))  /* An integer with one bit set. */

/* Typedefs that define the bitmap type for searching (EBITMAP),
 * the meta-character structure for MAGIC mode searching (MC),
 * and the meta-character structure for MAGIC mode replacment (RMC).
 */
typedef char *EBITMAP;

typedef struct
{
  int mc_type;
  union
  {
    int lchar;
    int group_no;
    EBITMAP cclmap;
  } u;
} MC;

typedef struct
{
  int mc_type;
  union
  {
    int group_no;
    char *rstr;
  } u;
} RMC;


/*  structure to hold user variables and their definitions  */

typedef struct UVAR
{
  char u_name[NVSIZE + 1];  /* name of user variable */
  char *u_value;            /* value (string) */
} UVAR;

/*  current user variables (This structure will probably change)  */

#define MAXUVARS   512

UVAR uv[MAXUVARS + 1];   /* user variables */

/* struct for ufuncs */

typedef struct UFUNC
{
  char *f_name;     /* name of function */
  int f_type;       /* 1 = monamic, 2 = dynamic */
} UFUNC;

/* defines for ufuncs */

#define NILNAMIC  0
#define MONAMIC   1
#define DYNAMIC   2
#define TRINAMIC  3

/******************************* ENVARS **********************************/


#define EVASAVE      0
#define EVBANGLINE   1
#define EVBUFHOOK    2
#define EVCFNAME     3
#define EVCMDHK      4
#define EVCOMPILER   5
#define EVDEBUG      6
#define EVEXBHOOK    7
#define EVFILLCOL    8
#define EVFMTOPTS    9
#define EVHARDTAB   10
#define EVREADHK    11
#define EVREPLACE   12
#define EVRVAL      13
#define EVSEARCH    14
#define EVSOFTTAB   15
#define EVSORTOPTS  16
#define EVSTATUS    17
#define EVTAGCMD    18
#define EVTIME      19
#define EVWCHARS    20
#define EVWRAPHK    21
#define EVWRITEHK   22

/******************************* UFUNCS **********************************/

#define UFABS      0
#define UFADD      1
#define UFAND      2
#define UFBIND     3
#define UFCAT      4
#define UFCHR      5
#define UFDIV      6
#define UFENV      7
#define UFEQUAL    8
#define UFEXIST    9
#define UFFIND    10
#define UFGREATER 11
#define UFGTKEY   12
#define UFIND     13
#define UFLEFT    14
#define UFLESS    15
#define UFLOWER   16
#define UFMID     17
#define UFMOD     18
#define UFNEG     19
#define UFNOT     20
#define UFOR      21
#define UFRIGHT   22
#define UFSEQUAL  23
#define UFSGREAT  24
#define UFSLESS   25
#define UFSUB     26
#define UFTIMES   27
#define UFUPPER   28




/* UNINITIALIZED GLOBAL DEFINITIONS */



BUFFER *bheadp;           /* Head of list of buffers      */
BUFFER *blistp;           /* Buffer for C-X C-B           */
BUFFER *curbp;            /* Current buffer               */

int curcol;               /* Cursor column                */
int curgoal;              /* Goal for C-P, C-N            */
int currow;               /* Cursor row                   */
int lastflag;             /* Flags, last command          */
int thisflag;             /* Flags, this command          */
SCRN *first_screen;     /* Head and current screen in list */
unsigned char pat[NPAT];  /* Search pattern               */
unsigned char rpat[NPAT]; /* replacement pattern          */
unsigned char tap[NPAT];  /* Reversed pattern array.      */
WNDW *curwp;            /* Current window               */
WNDW *wheadp;           /* Head of list of windows      */
extern TERM term;         /* Terminal information.        */

/*  Various "Hook" execution variables  */

KEYTAB bufhook;     /* executed when entering a buffer */
KEYTAB cmdhook;     /* executed before looking for a command */
KEYTAB exbhook;     /* executed when exiting a buffer */
KEYTAB readhook;    /* executed on all file reads */
KEYTAB wraphook;    /* executed when wrapping text */
KEYTAB writehook;   /* executed on all file writes */

/* The variables matchline and matchoff hold the line
 * and offset position of the *start* of the match.
 * The variable patmatch holds the string that satisfies
 * the search command.
 */

int matchlen;
int matchoff;
LINE *matchline;

char *grpmatch[MAXGROUPS];  /* holds groups found in search */
MC mcpat[NPAT];     /* the magic pattern            */
MC tapcm[NPAT];     /* the reversed magic pattern   */
RMC rmcpat[NPAT];   /* the replacement magic array  */


/* VARIABLES INITIALIZED IN RHO_INIT.H */

const extern char *modename[];
const extern char modecode[];
const extern char *search_path[];
extern char *envars[];
extern char *execstr;
extern char *fline;
extern char *lastptr;
extern char *patmatch;
extern char bangline[];
extern char compiler[];
extern char fmtlead[];
extern char fmtopts[];
extern char golabel[];
extern char lastmesg[];
extern char lterm[];
extern char mainbuf[LENPATH];
extern char paralead[];
extern char sortopts[];
extern char tagcmd[];
extern const char *dname[];
extern const char errorm[];
extern const char falsem[];
extern const char truem[];
extern const int dname_len[NUMDIRS];
extern int *kbdend;
extern int *kbdptr;
extern int abortc;
extern int asave;
extern int asavecount;
extern int charpending;
extern int clexec;
extern int cmdstatus;
extern int cpending;
extern int debug;
extern int diagflag;
extern int discmd;
extern int disinp;
extern int eexitflag;
extern int eolexist;
extern int execlevel;
extern int exec_error;
extern int fillcol;
extern int gflags;
extern int gmode;
extern int hardtab;
extern int hjump;
extern int hscroll;
extern int hscrollbar;
extern int isterm;
extern int kbdmode;
extern int kbdm[];
extern int kbdrep;
extern int kill_index;
extern int kskip[];
extern int kused[];
extern int lastkey;
extern int lbound;
extern int magical;
extern int modeflag;
extern int mpresf;
extern int mstore;
extern int nenvars;
extern int npnames;
extern int numfunc;
extern int overlap;
extern int posflag;
extern int predef;
extern int prefix;
extern int prenum;
extern int quotec;
extern int reptc;
extern int revexist;
extern int rmagical;
extern int rval;
extern int saveflag;
extern int searchtype;
extern int sgarbf;
extern int softtab;
extern int sourcing;
extern int sscroll;
extern int sterm;
extern int taboff;
extern int ttcol;
extern int ttrow;
extern int vscrollbar;
extern int vtcol;
extern int vtrow;
extern int wlflag;
extern int xpos;
extern int yankflag;
extern int ypos;
extern KEYTAB keytab[];
extern KILL *kbufh[];
extern KILL *kbufp[];
extern NBIND names[];
extern size_t flen;
extern struct BUFFER *bstore;
extern unsigned char wordlist[];
extern WNDW *swindow;

/* RHO.C EXTERNS FOR RHO*.H */

extern int delgmode (int f, int n);
extern int delmode (int f, int n);
extern int exit_rho (int f, int n);
extern int exit_status (int);
extern int setgmode (int f, int n);
extern int setmod (int f, int n);
extern int update_screen (int f, int n);

